/*
 * Copyright (C) 1997-2005 Andre Noll <maan@systemlinux.org>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "para.h"
#include "config.h"
#include <readline/readline.h>
#include <readline/history.h>
#include "client.cmdline.h"

/* A static variable for holding the line. */
static char *line_read = (char *)NULL;

struct gengetopt_args_info args_info;

/*
 * client log function
 */
void clog(int, char*, ...) __attribute__((format (printf, 2, 3)));
void clog(int ll, char* fmt,...)
{
	va_list argp;
	FILE *outfd;

	/* ignore log message if loglevel is not high enough */
	if (ll < args_info.loglevel_arg)
		return;
	if (ll < WARNING)
		outfd = stdout;
	else
		outfd = stderr;
	va_start(argp, fmt);
	vfprintf(stdout, fmt, argp);
	va_end(argp);
}
/*
 * decrypts challgenge number with rsa key of user. decrypted number is stored
 * in challenge_nr. Returns 0 on success.
 */
int decrypt(char *seckey_file, long int *challenge_nr, char *buf,
		int numbytes)
{
	int n, fd1[2], fd2[2];
	pid_t pid;

	if(pipe(fd1) < 0 || pipe(fd2) < 0)
		perror("pipe");
	if ((pid = fork()) < 0) {
		perror("fork");
	} else if (pid > 0) {
		/* parent */
		char *line = buf;
		int num;
		close(fd1[0]);
		close(fd2[1]);
		clog(DEBUG, "%s: writing encrypted challenge to pipe "
			"(%i bytes)\n", __func__, numbytes);
		if ((num = write(fd1[1], line, numbytes)) != numbytes) {
			clog(ERROR, "%s: write error to pipe "
				"(wrote only %i)\n", __func__, num);
			return -1;
		}
		close(fd1[1]);
		if ((n = read(fd2[0], line, MAXLINE)) < 0) {
			clog(ERROR, "%s: read error from pipe\n", __func__);
			return -1;
		}
		line[n] = '\0';
	 	close(fd2[0]);
		if (n <= 0)
			return -1;
		if (sscanf(line, "%lu", challenge_nr) <= 0)
			return -1;
		return 0;
	} else { /* child*/
		close(fd1[1]);
		close(fd2[0]);
		if (fd1[0] != STDIN_FILENO) {
			if (dup2(fd1[0], STDIN_FILENO) != STDIN_FILENO) {
				clog(ERROR, "%s: dup2 error to stdin\n",
					__func__);
				close(fd1[0]);
			}
		}
		if (fd2[1] != STDOUT_FILENO) {
			if (dup2(fd2[1], STDOUT_FILENO) != STDOUT_FILENO ) {
				clog(ERROR, "%s: dup2 error to stdout\n",
					__func__);
				close(fd2[1]);
			}
		}
		if (execl(OPENSSL, "openssl", "rsautl", "-decrypt", "-inkey",
				seckey_file, NULL) < 0) {
			clog(ERROR, "%s: execl error\n", __func__);
			exit(EXIT_FAILURE);
		}
	}
	return -1; /* never reached*/
}

/*
 * Read a string, and return a pointer to it. Returns NULL on EOF.
 */
char *rl_gets()
{
	/* If the buffer has already been allocated, return the memory
	 * to the free pool.
	 */
	if (line_read) {
		free (line_read);
		line_read = (char *)NULL;
	}
	/* Get a line from the user. */
	line_read = readline ("para_client> ");
	/* If the line has any text in it, save it on the history. */
	if (line_read && *line_read)
		add_history (line_read);
	return line_read;
}

/*
 * do several cleanups on sigint
 */
void signals_int(int i)
{
	rl_save_prompt();
	printf("\n");
	rl_message("\n");
	rl_kill_full_line(0, 0);
	rl_restore_prompt();
	rl_free_line_state();
	rl_cleanup_after_signal();
	rl_reset_after_signal();
	//printf("Caught SIGINT\n");
}

/*
 * return proper exit status
 */
int get_exit_status (int status)
{
	if (WIFEXITED(status))
		return WEXITSTATUS(status);
	if (WIFSIGNALED(status))
		return WTERMSIG(status);
	return WSTOPSIG(status);
}

/*
 * wait and tell exit status
 */
void sigchld_handler(int s)
{
	int i,es,logl;
	pid_t pid;
	
	pid = wait(&i);
	es = get_exit_status(i);
	//printf("verbose=%i\n",verbose);
	if (es)
		logl = ERROR;
	else
		logl = DEBUG;
	clog(logl, "%s: child process %i died. Exit status: %i (%s)\n",
		__func__, pid, es, strerror(es));
}

void get_options(int argc, char *argv[],
	char **config_file, char **key_file)
{
	char *home;
	static char default_key_file[_POSIX_PATH_MAX] = "";
	static char default_config_file[_POSIX_PATH_MAX] = "";
	struct stat statbuf;
	int ret;

	cmdline_parser(argc, argv, &args_info);
	if (!args_info.user_given)
		args_info.user_arg = para_logname();
	if (!args_info.key_file_given) {
		home = para_homedir();
		sprintf(default_key_file, "%s/.paraslash/key.%s", home, 
			args_info.user_arg);
		free(home);
	}
	if (!args_info.config_file_given) {
		home = para_homedir();
		sprintf(default_config_file, "%s/.paraslash/client.conf",
			home);
		free(home);
	}
	if (!args_info.config_file_given)
		*config_file = default_config_file;
	else
		*config_file = args_info.config_file_arg;
	ret = stat(*config_file, &statbuf);
	if (ret && args_info.config_file_given) {
		fprintf(stderr, "can not stat config file %s\n",
			args_info.config_file_arg);
		exit(EXIT_FAILURE);
	}
	if (!ret)
		cmdline_parser_configfile(*config_file, &args_info, 0, 0, 0);
	if (!args_info.key_file_given)
		*key_file = default_key_file;
	else
		*key_file = args_info.key_file_arg;
}

/*
 * MAIN
 */
int main(int argc, char *argv[]) {

	struct sigaction sa, sa_chld;
	int sockfd, numbytes, i, interactive;
	struct hostent *he;
	struct sockaddr_in their_addr;
	char *command = NULL;
	char buf[STRINGSIZE];
	char challenge_decrypt[STRINGSIZE];
	char auth_str[MAXLINE]; /* FIXME: use send_va_buffer instead */
	char *key_file, *config_file;
	long int challenge_nr;
	char *line;

	get_options(argc, argv, &config_file, &key_file);
	if (args_info.loglevel_arg <= NOTICE)
		cmdline_parser_print_version();
	clog(INFO, "%s:"
		"current loglevel: %d\n"
		"using config_file: %s\n"
		"using key_file: %s\n"
		"connecting to %s:%d\n",
		__func__,
		args_info.loglevel_arg,
		config_file,
		key_file,
		args_info.hostname_arg,
		args_info.server_port_arg
	);
	
	//setlinebuf(stdout);
	/* compute auth string to be sent to server */
	sprintf(auth_str, "auth %s", args_info.user_arg);
	/* install sigchld handler */
	sa_chld.sa_handler = sigchld_handler;
	sigemptyset(&sa_chld.sa_mask);
	sa_chld.sa_flags = SA_RESTART;
	if (sigaction(SIGCHLD, &sa_chld, NULL) == -1) {
		clog(EMERG, "%s: failed to install sigchld handler\n", 
			__func__);
		exit(EXIT_FAILURE);
	}
	interactive = args_info.inputs_num == 0? 1 : 0;
	if (interactive) {
		/* catch SIGINT only if in interactive mode */
		sa.sa_flags = 0;
		sa.sa_handler = signals_int;
		sigaction(SIGINT, &sa, NULL);
		clog(NOTICE, "%s: no command given, entering interactive "
			"mode\n", __func__);
	} else {
		/* not interactive, copy command and args to command */
		for (i = 0; i < args_info.inputs_num; i++) {
			size_t len = strlen(args_info.inputs[i]);
			if (command) {
				command = realloc(command, strlen(command) 
					+ len + 2);
				strcat(command, "\n");
			} else {
				command = malloc((len + 2) * sizeof(char));
				*command = '\0';
			}
			strcat(command, args_info.inputs[i]);
		}
		//clog(NOTICE, "%s: Using command: %s\n", __func__, command);
	}
	//exit(1);
interactive_loop:
	if (interactive) {
		int i = 0;
		char *p;

		/* read a line via readline */
		line = rl_gets();
		if (!line)
			return 0;
		if (!line[0])
			goto interactive_loop;
		p = line;
		while (sscanf(p, "%200s%n", buf, &i) == 1) {
			size_t len = strlen(buf);
			if (command) {
				command = realloc(command, strlen(command) 
					+ len + 2);
				strcat(command, "\n");
			} else {
				command = malloc((len + 2) * sizeof(char));
				*command = '\0';
			}
			strcat(command, buf);
			p += i;
		}
//		printf("%s\n", command);
	}
	/* get the host info */
	clog(NOTICE, "%s: getting host info of %s\n", __func__,
		args_info.hostname_arg);
	if (!(he = get_host_info(args_info.hostname_arg)))
		exit(EXIT_FAILURE);
	/* get new socket */
	if ((sockfd = get_socket()) < 0)
		exit(EXIT_FAILURE);
	/* init their_addr */
	init_sockaddr(&their_addr, args_info.server_port_arg, he);
	/* Connect */
	clog(NOTICE, "%s: connecting to %s...\n", __func__,
		args_info.hostname_arg);
	if (do_connect(sockfd, &their_addr) < 0)
		exit(EXIT_FAILURE);
	/* Receive Welcome message */
	if ((numbytes = recv_buffer(sockfd, buf, MAXLINE)) < 0)
		exit(EXIT_FAILURE);
	/* send auth command */
	clog(INFO, "%s: <-- %s--> %s\n", __func__, buf, auth_str);
	if (send_buffer(sockfd, auth_str) < 0)
		exit(EXIT_FAILURE);
	/* receive challenge number */
	if ((numbytes = recv_buffer(sockfd, buf, MAXLINE)) < 0)
		exit(EXIT_FAILURE);
	if (numbytes != 64) {
		clog(EMERG, "%s: did not receive valid challenge (got %i bytes)\n", 
			__func__, numbytes);
		buf[numbytes] = '\0';
		clog(ERROR, "%s: Received the following instead: %s\n", 
			__func__, buf);
		goto write_out;
	}
	clog(INFO, "%s: <-- [challenge (%i bytes)]\n", __func__, numbytes);
	/* decrypt challenge number */
	if (decrypt(key_file, &challenge_nr, buf, numbytes) < 0) {
		clog(EMERG, "%s: decrypt error. Bad secret key?\n", __func__);
		exit(EXIT_FAILURE);
	}
	/* send decrypted challenge */
	sprintf(challenge_decrypt, "challenge_response:%lu", challenge_nr);
	clog(INFO, "%s: --> %lu\n", __func__, challenge_nr);
	if (send_buffer(sockfd, challenge_decrypt) < 0)
		exit(EXIT_FAILURE);
	/* Wait for approval */
	clog(NOTICE, "%s: waiting for approval from server\n", __func__);
	if ((numbytes = recv_buffer(sockfd, buf, MAXLINE)) < 0)
		exit(EXIT_FAILURE);
	clog(INFO, "%s: ++++ server info ++++\n%s\n++++ end of server "
		"info ++++\n", __func__, buf);
	/* Check if server has sent "Proceed" message */
	if (!strstr(buf, PROCEED_MSG)) {
		clog(EMERG, "%s: authentication failed\n", __func__);
		exit(EXIT_FAILURE);
	}
	/* send command */
	clog(INFO, "%s: --> %s\n", __func__, command);
	if (send_buffer(sockfd, command) < 0)
		exit(EXIT_FAILURE);
	free(command);
	command = NULL;
	if (send_buffer(sockfd, EOC_MSG "\n") < 0)
		exit(EXIT_FAILURE);
	clog(NOTICE, "%s: command sent.\n", __func__);
write_out:
	/* write server output to stdout */
	while ((numbytes = recv_buffer(sockfd, buf, STRINGSIZE)) > 0) {
		int ret;

		if (strstr(buf, AWAITING_DATA_MSG)) {
			clog(NOTICE, "%s: <-- awaiting data\n", __func__);
			clog(NOTICE, "%s: --> sending stdin\n", __func__);
			while ((ret = read(STDIN_FILENO, buf,
					sizeof(buf) - 1)) > 0)
				send_bin_buffer(sockfd, buf, ret);
			clog(NOTICE, "%s: closing connection\n", __func__);
			numbytes = 1;
			break;
		}
		if (write(STDOUT_FILENO, buf, numbytes) != numbytes)
			exit(EXIT_FAILURE);
	//	sleep(2);
	}
	if (numbytes < 0)
		exit(EXIT_FAILURE);
	if (!numbytes)
		clog(NOTICE, "%s: connection closed by peer\n", __func__);
	close(sockfd);
	if (interactive)
		goto interactive_loop;
	return 0;
}
